/* smbget: a wget-like utility with support for recursive downloading and 
 * smb:// urls */

/* (C) 2003 Jelmer Vernooij <jelmer@samba.org> under the GPL  */

#define VERSION "0.1"

#define _GNU_SOURCE
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <popt.h>
#include <stdlib.h>
#include <libsmbclient.h>
#include <errno.h>

/* FIXME */
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

#define SMB_MAXPATHLEN MAXPATHLEN
/* Number of bytes to read when checking whether local and remote file are really the same file */
#define RESUME_CHECK_SIZE 				512
#define RESUME_DOWNLOAD_OFFSET			1024
#define RESUME_CHECK_OFFSET				RESUME_DOWNLOAD_OFFSET+RESUME_CHECK_SIZE
/* Number of bytes to read at once */
#define SMB_BLOCKSIZE 					640

char *username = NULL, *password = NULL, *workgroup = NULL;
int nonprompt = 0, quiet = 0, dots = 0;

int smb_download_file(const char *base, char *name, int recursive, int resume);

void get_auth_data(const char *srv, const char *shr, char *wg, int wglen, char *un, int unlen, char *pw, int pwlen)
{
	char *wgtmp, *usertmp;
	char tmp[128];

	if(!nonprompt && !username) {
		printf("Username for %s at %s [%s] ", shr, srv, un);
		fgets(tmp, sizeof(tmp), stdin);
		if(tmp[strlen(tmp)-1] == '\n')tmp[strlen(tmp)-1] = '\0';
		strncpy(un, tmp, unlen-1);
	} else if(username) strncpy(un, username, unlen-1);

	if(!nonprompt && !password) {
		printf("Password for %s at %s: ", shr, srv);
		fgets(tmp, sizeof(tmp), stdin);
		if(tmp[strlen(tmp)-1] == '\n')tmp[strlen(tmp)-1] = '\0';
		strncpy(pw, tmp, pwlen-1);
	} else if(password) strncpy(pw, password, pwlen-1);


	wgtmp = strndup(wg, wglen); 
	usertmp = strndup(un, unlen);
	printf("Using workgroup %s, %s%s\n", wgtmp, *usertmp?"user ":"guest user", usertmp);
	free(wgtmp); free(usertmp);
}

int smb_download_dir(const char *base, char *name, int resume)
{
	char path[SMB_MAXPATHLEN];
	snprintf(path, SMB_MAXPATHLEN-1, "%s%s%s", base, (*base && *name)?"/":"", name);
	int dirhandle;
	struct smbc_dirent *dirent;

	mkdir(name, 0666);

	/* List files in directory and call smb_download_file on them */
	dirhandle = smbc_opendir(path);
	if(dirhandle < 1) {
		fprintf(stderr, "Can't open directory %s: %s\n", path, strerror(errno));
		return 0;
	}

	while((dirent = smbc_readdir(dirhandle))) {
		char *newname;
		asprintf(&newname, "%s/%s", name, dirent->name);
		switch(dirent->smbc_type) {
		case SMBC_DIR:
			smb_download_dir(base, newname, resume);
			break;

		case SMBC_FILE:
			smb_download_file(base, newname, 1, resume);
			break;

		default:
			break;
		}
		free(newname);
	}

	smbc_closedir(dirhandle);
	return 1;
}

int smb_download_file(const char *base, char *name, int recursive, int resume) {
	int remotehandle, localhandle;
	const char *newpath;
	char path[SMB_MAXPATHLEN];
	char checkbuf[2][RESUME_CHECK_SIZE];
	char readbuf[SMB_BLOCKSIZE];
	off_t offset_download = 0, offset_check = 0, curpos = 0;
	struct stat localstat, remotestat;

	snprintf(path, SMB_MAXPATHLEN-1, "%s%s%s", base, (*base && *name)?"/":"", name);
	
	remotehandle = smbc_open(path, O_RDONLY, 0755);

	if(remotehandle < 0) {
		switch(errno) {
		case EISDIR: 
			if(!recursive) {
				fprintf(stderr, "%s is a directory. Specify -R to download recursively\n", path);
				return 0;
			}
			smb_download_dir(base, name, resume);
			return 0;

		case ENOENT:
			fprintf(stderr, "%s can't be found on the remote server\n", path);
			return 0;

		case ENOMEM:
			fprintf(stderr, "Not enough memory\n");
			exit(1);
			return 0;

		case ENODEV:
			fprintf(stderr, "The share name used in %s does not exist\n", path);
			return 0;

		case EACCES:
			fprintf(stderr, "You don't have enough permissions to access %s\n", path);
			return 0;

		default:
			perror("smbc_open");
			return 0;
		}
	} 

	if(smbc_fstat(remotehandle, &remotestat) < 0) {
		fprintf(stderr, "Can't stat %s: %s\n", path, strerror(errno));
		return 0;
	}

	if(!quiet)printf("Remote file size: %lu\n", remotestat.st_size);

	if(!name[0]) {
		newpath = strrchr(base, '/');
		if(newpath)newpath++; else newpath = base;
	} else newpath = name;
	
	/* Open local file and, if necessary, resume */
	localhandle = open(newpath, O_CREAT | O_NONBLOCK | O_RDWR | (!resume?O_EXCL:0), 0755);
	if(localhandle < 0) {
		fprintf(stderr, "Can't open %s: %s\n", newpath, strerror(errno));
		smbc_close(remotehandle);
		return 0;
	}
	
	fstat(localhandle, &localstat);

	if(localstat.st_size && localstat.st_size == remotestat.st_size) {
		fprintf(stderr, "%s is already downloaded completely.\n", path);
		return 1;
	}

	if(localstat.st_size > RESUME_CHECK_OFFSET && remotestat.st_size > RESUME_CHECK_OFFSET) {
		offset_download = localstat.st_size - RESUME_DOWNLOAD_OFFSET;
		offset_check = localstat.st_size - RESUME_CHECK_OFFSET;
		printf("Trying to start resume of %s at %lu\n"
			   "At the moment %lu of %lu bytes have been retrieved\n", newpath, offset_check, 
			   localstat.st_size, remotestat.st_size);
	}

	if(offset_check) { 
		off_t off1, off2;
		/* First, check all bytes from offset_check to offset_download */
		off1 = lseek(localhandle, offset_check, SEEK_SET);
		if(off1 < 0) {
			fprintf(stderr, "Can't seek to %lu in local file %s\n", offset_check, newpath);
			smbc_close(remotehandle); close(localhandle);
			return 0;
		}

		off2 = smbc_lseek(remotehandle, offset_check, SEEK_SET); 
		if(off2 < 0) {
			fprintf(stderr, "Can't seek to %lu in remote file %s\n", offset_check, newpath);
			smbc_close(remotehandle); close(localhandle);
			return 0;
		}

		if(off1 != off2) {
			fprintf(stderr, "Offset in local and remote files is different (local: %lu, remote: %lu)\n", off1, off2);
			return 0;
		}

		if(smbc_read(remotehandle, checkbuf[0], RESUME_CHECK_SIZE) != RESUME_CHECK_SIZE) {
			fprintf(stderr, "Can't read %d bytes from remote file %s\n", RESUME_CHECK_SIZE, path);
			smbc_close(remotehandle); close(localhandle);
			return 0;
		}

		if(read(localhandle, checkbuf[1], RESUME_CHECK_SIZE) != RESUME_CHECK_SIZE) {
			fprintf(stderr, "Can't read %d bytes from local file %s\n", RESUME_CHECK_SIZE, name);
			smbc_close(remotehandle); close(localhandle);
			return 0;
		}

		if(memcmp(checkbuf[0], checkbuf[1], RESUME_CHECK_SIZE) == 0) {
			printf("Current local and remote file appear to be the same. Starting download from offset %ld\n", offset_download);
		} else {
			fprintf(stderr, "Local and remote file appear to be different, not doing resume for %s\n", path);
//			return 0;
		}
	}

	/* Now, download all bytes from offset_download to the end */
	for(curpos = offset_download; curpos < remotestat.st_size; curpos+=SMB_BLOCKSIZE) {
		ssize_t bytesread = smbc_read(remotehandle, readbuf, SMB_BLOCKSIZE);
		if(bytesread < 0) {
			fprintf(stderr, "Can't read %d bytes at offset %lu, file %s\n", SMB_BLOCKSIZE, curpos, path);
			smbc_close(remotehandle); close(localhandle);
			return 0;
		}

		if(write(localhandle, readbuf, bytesread) < 0) {
			fprintf(stderr, "Can't write %d bytes to local file %s at offset %lu\n", bytesread, path, curpos);
			return 0;
		}
		if(dots)fputc('.', stderr);
		else if(!quiet) {
			
			printf("%d bytes downloaded\n", bytesread);
		}
	}
	if(dots)fputc('\n', stderr);

	printf("%s downloaded\n", path);

	smbc_close(remotehandle);
	close(localhandle);

	return 1;
}

int smb_download(const char *base, char *name, int recursive, int resume) {
	if(!recursive) return smb_download_file(base, name, recursive, resume);
	else return smb_download_file(base, name, recursive, resume);
}
                                                                                
int main(int argc, const char **argv)
{
	int resume = 0, recursive = 0;
	int c = 0;
	int debuglevel = 0;
	const char *file = NULL;
	struct poptOption long_options[] = {
		{"guest", 'a', POPT_ARG_NONE, NULL, 'a', "Work as user guest" },	
		{"resume", 'r', POPT_ARG_NONE, &resume, 0, "Automatically resume aborted files" },
		{"recursive", 'R',  POPT_ARG_NONE, &recursive, 0, "Recursively download files" },
		{"username", 'u', POPT_ARG_STRING, &username, 'u', "Username to use" },
		{"password", 'p', POPT_ARG_STRING, &password, 'p', "Password to use" },
		{"workgroup", 'w', POPT_ARG_STRING, &workgroup, 'w', "Workgroup to use (optional)" },
		{"nonprompt", 'n', POPT_ARG_NONE, &nonprompt, 'n', "Don't ask anything (non-interactive)" },
		{"debuglevel", 'd', POPT_ARG_INT, &debuglevel, 'd', "Debuglevel to use" },
		{"dots", 'D', POPT_ARG_NONE, &dots, 'D', "Show dots as progress indication" },
		{"quiet", 'q', POPT_ARG_NONE, &quiet, 'q', "Be quiet" },
		POPT_AUTOHELP
		POPT_TABLEEND
	};
	poptContext pc;

	pc = poptGetContext(argv[0], argc, argv, long_options, 0);

	while((c = poptGetNextOpt(pc)) >= 0) {
		switch(c) {
		case 'a':
			username = ""; password = "";
			printf("Working as guest user\n");
			break;
		}
	}

	printf("SMBGet Version "VERSION"\n");
	printf("(c) 2003 Jelmer Vernooij <jelmer@samba.org>\n");

	if(smbc_init(get_auth_data, debuglevel) < 0) {
		fprintf(stderr, "Unable to initialize libsmbclient\n");
		return 1;
	}

	while((file = poptGetArg(pc))) {
		smb_download(file, "", recursive, resume);
	}

	return 0;
}
